<?php

// autoload_static.php @generated by Composer

namespace Composer\Autoload;

class ComposerStaticInitPhar
{
    public static $prefixLengthsPsr4 = array (
        'S' =>
        array (
            'Sit\\' => 4,
        ),
        'Q' =>
        array (
            'Qux\\' => 4,
        ),
        'D' =>
        array (
            'Dolor\\' => 6,
        ),
        'B' =>
        array (
            'Baz\\' => 4,
        ),
    );

    public static $prefixDirsPsr4 = array (
        'Sit\\' =>
        array (
            0 => 'phar://' . __DIR__ . '/..' . '/a/a/dir/sit.phar/src',
        ),
        'Qux\\' =>
        array (
            0 => 'phar://' . __DIR__ . '/../..' . '/dir/qux.phar/src',
        ),
        'Dolor\\' =>
        array (
            0 => 'phar://' . __DIR__ . '/..' . '/a/a/dolor.phar',
        ),
        'Baz\\' =>
        array (
            0 => 'phar://' . __DIR__ . '/../..' . '/baz.phar',
        ),
    );

    public static $prefixesPsr0 = array (
        'L' =>
        array (
            'Lorem' =>
            array (
                0 => 'phar://' . __DIR__ . '/..' . '/a/a/lorem.phar',
            ),
        ),
        'I' =>
        array (
            'Ipsum' =>
            array (
                0 => 'phar://' . __DIR__ . '/..' . '/a/a/dir/ipsum.phar/src',
            ),
        ),
        'F' =>
        array (
            'Foo' =>
            array (
                0 => 'phar://' . __DIR__ . '/../..' . '/foo.phar',
            ),
        ),
        'B' =>
        array (
            'Bar' =>
            array (
                0 => 'phar://' . __DIR__ . '/../..' . '/dir/bar.phar/src',
            ),
        ),
    );

    public static $classMap = array (
        'Composer\\InstalledVersions' => __DIR__ . '/..' . '/composer/InstalledVersions.php',
    );

    public static function getInitializer(ClassLoader $loader)
    {
        return \Closure::bind(function () use ($loader) {
            $loader->prefixLengthsPsr4 = ComposerStaticInitPhar::$prefixLengthsPsr4;
            $loader->prefixDirsPsr4 = ComposerStaticInitPhar::$prefixDirsPsr4;
            $loader->prefixesPsr0 = ComposerStaticInitPhar::$prefixesPsr0;
            $loader->classMap = ComposerStaticInitPhar::$classMap;

        }, null, ClassLoader::class);
    }
}
