<?php

// autoload_classmap.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'A' => $vendorDir . '/a/a/src/A.php',
    'C' => $vendorDir . '/c/c/src/C.php',
    'Composer\\InstalledVersions' => $vendorDir . '/composer/InstalledVersions.php',
    'D' => $vendorDir . '/d/d/src/D.php',
);
