<?php

// autoload_namespaces.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'Main' => array($baseDir . '/src'),
    'Lala' => array($baseDir . '/src', $baseDir . '/lib'),
);
