<?php

// autoload_classmap.php @generated by Composer

$vendorDir = dirname(dirname(__FILE__));
$baseDir = dirname($vendorDir);

return array(
    'Composer\\InstalledVersions' => $vendorDir . '/composer/InstalledVersions.php',
    'Hooks' => $baseDir . '/Hooks.php',
    'PluginA' => $vendorDir . '/plugin/a/PluginA.php',
    'PluginB' => $vendorDir . '/plugin/b/PluginB.php',
);
