<?php

// autoload_files.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'fd1502a3b53705310b21799d025f0dee' => $baseDir . '/devfiles/foo.php',
);
