<?php

// autoload_classmap.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'ClassMapBar' => $vendorDir . '/b/b/test.php',
    'ClassMapBaz' => $vendorDir . '/c/c/foo/test.php',
    'ClassMapFoo' => $vendorDir . '/a/a/src/a.php',
    'Composer\\InstalledVersions' => $vendorDir . '/composer/InstalledVersions.php',
);
