<?php

// autoload_classmap.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = $vendorDir;

return array(
    'ClassMapFoo' => $vendorDir . '/composersrc/foo.php',
    'Composer\\InstalledVersions' => $vendorDir . '/composer/InstalledVersions.php',
    'Main\\Foo' => $vendorDir . '/src/Main/Foo.php',
);
