<?php

// autoload_files.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    $vendorDir . '/c/lorem/testC.php',
    $vendorDir . '/z/foo/testA.php',
    $vendorDir . '/d/d/testD.php',
    $vendorDir . '/b/bar/testB.php',
    $vendorDir . '/e/e/testE.php',
    $baseDir . '/root.php',
);
