<?php

// autoload_psr4.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'Sit\\' => array('phar://' . $vendorDir . '/a/a/dir/sit.phar/src'),
    'Qux\\' => array('phar://' . $baseDir . '/dir/qux.phar/src'),
    'Dolor\\' => array('phar://' . $vendorDir . '/a/a/dolor.phar'),
    'Baz\\' => array('phar://' . $baseDir . '/baz.phar'),
);
