<?php

// autoload_static.php @generated by Composer

namespace Composer\Autoload;

class ComposerStaticInitTargetDir
{
    public static $files = array (
        'b4b4f2eb2151c57cadbd5714a7aba8b7' => __DIR__ . '/../..' . '/foo.php',
        '99b24fc198db06c1d2d8118a8a5475eb' => __DIR__ . '/../..' . '/bar.php',
    );

    public static $prefixesPsr0 = array (
        'M' =>
        array (
            'Main\\Foo' =>
            array (
                0 => __DIR__ . '/../..' . '/',
            ),
            'Main\\Bar' =>
            array (
                0 => __DIR__ . '/../..' . '/',
            ),
        ),
    );

    public static $classMap = array (
        'ClassMapBar' => __DIR__ . '/../..' . '/lib/rootbar.php',
        'ClassMapFoo' => __DIR__ . '/../..' . '/src/rootfoo.php',
        'Composer\\InstalledVersions' => __DIR__ . '/..' . '/composer/InstalledVersions.php',
    );

    public static function getInitializer(ClassLoader $loader)
    {
        return \Closure::bind(function () use ($loader) {
            $loader->prefixesPsr0 = ComposerStaticInitTargetDir::$prefixesPsr0;
            $loader->classMap = ComposerStaticInitTargetDir::$classMap;

        }, null, ClassLoader::class);
    }
}
