<?php

// autoload_namespaces.php @generated by Composer

$vendorDir = dirname(__DIR__);
$baseDir = dirname($vendorDir);

return array(
    'Lorem' => array('phar://' . $vendorDir . '/a/a/lorem.phar'),
    'Ipsum' => array('phar://' . $vendorDir . '/a/a/dir/ipsum.phar/src'),
    'Foo' => array('phar://' . $baseDir . '/foo.phar'),
    'Bar' => array('phar://' . $baseDir . '/dir/bar.phar/src'),
);
